@@ -291,6 +291,9 @@ class AxiMasterWrite(Region, Reset):
         if isinstance(data, int):
             raise ValueError("Expected bytes or bytearray for data")
 
+        if burst != AxiBurstType.FIXED and address+len(data) > 2**self.address_width:
+            raise ValueError("Requested transfer overruns end of address space")
+
         if awid is None or awid < 0:
             awid = None
         elif awid > self.id_count:
@@ -357,6 +360,9 @@ class AxiMasterWrite(Region, Reset):
         if isinstance(data, int):
             raise ValueError("Expected bytes or bytearray for data")
 
+        if burst != AxiBurstType.FIXED and address+len(data) > 2**self.address_width:
+            raise ValueError("Requested transfer overruns end of address space")
+
         if awid is None or awid < 0:
             awid = None
         elif awid > self.id_count:
@@ -476,7 +482,7 @@ class AxiMasterWrite(Region, Reset):
 
             cycles = (len(cmd.data) + (cmd.address % num_bytes) + num_bytes-1) // num_bytes
 
-            cur_addr = aligned_addr
+            cur_addr = cmd.address
             offset = 0
             cycle_offset = aligned_addr-word_addr
             n = 0
@@ -563,7 +569,12 @@ class AxiMasterWrite(Region, Reset):
 
                 await self.w_channel.send(w)
 
-                cur_addr += num_bytes
+                if cmd.burst == AxiBurstType.FIXED:
+                    cur_addr = cmd.address
+                elif k == 0:
+                    cur_addr = aligned_addr + num_bytes
+                else:
+                    cur_addr += num_bytes
                 cycle_offset = (cycle_offset + num_bytes) % self.byte_lanes
 
             resp_cmd = AxiWriteRespCmd(cmd.address, len(cmd.data), cmd.size, cycles, cmd.prot, burst_list, cmd.event)
@@ -577,8 +588,7 @@ class AxiMasterWrite(Region, Reset):
 
             bid = int(getattr(b, 'bid', 0))
 
-            if self.active_id[bid] <= 0:
-                raise Exception(f"Unexpected burst ID {bid}")
+            assert self.active_id[bid] > 0, "unexpected burst ID"
 
             self.tag_context_manager.put_resp(bid, b)
 
@@ -591,7 +601,7 @@ class AxiMasterWrite(Region, Reset):
         for burst_length in cmd.burst_list:
             b = await context.get_resp()
 
-            burst_resp = AxiResp(getattr(b, 'bresp', AxiResp.OKAY))
+            burst_resp = AxiResp(int(getattr(b, 'bresp', AxiResp.OKAY)))
             burst_user = int(getattr(b, 'buser', 0))
 
             if burst_resp != AxiResp.OKAY:
@@ -600,8 +610,7 @@ class AxiMasterWrite(Region, Reset):
             if burst_user is not None:
                 user.append(burst_user)
 
-            if self.active_id[bid] <= 0:
-                raise Exception(f"Unexpected burst ID {bid}")
+            assert self.active_id[bid] > 0, "unexpected burst ID"
 
             self.active_id[bid] -= 1
 
@@ -714,6 +723,9 @@ class AxiMasterRead(Region, Reset):
         if length < 0:
             raise ValueError("Read length must be positive")
 
+        if burst != AxiBurstType.FIXED and address+length > 2**self.address_width:
+            raise ValueError("Requested transfer overruns end of address space")
+
         if arid is None or arid < 0:
             arid = None
         elif arid > self.id_count:
@@ -768,6 +780,9 @@ class AxiMasterRead(Region, Reset):
         if length < 0:
             raise ValueError("Read length must be positive")
 
+        if burst != AxiBurstType.FIXED and address+length > 2**self.address_width:
+            raise ValueError("Requested transfer overruns end of address space")
+
         if arid is None or arid < 0:
             arid = None
         elif arid > self.id_count:
@@ -872,7 +887,7 @@ class AxiMasterRead(Region, Reset):
 
             burst_list = []
 
-            cur_addr = aligned_addr
+            cur_addr = cmd.address
             n = 0
 
             burst_length = 0
@@ -917,7 +932,12 @@ class AxiMasterRead(Region, Reset):
                     self.log.info("Read burst start arid: 0x%x araddr: 0x%08x arlen: %d arsize: %d arprot: %s",
                             arid, cur_addr, burst_length-1, cmd.size, cmd.prot)
 
-                cur_addr += num_bytes
+                if cmd.burst == AxiBurstType.FIXED:
+                    cur_addr = cmd.address
+                elif k == 0:
+                    cur_addr = aligned_addr + num_bytes
+                else:
+                    cur_addr += num_bytes
 
             resp_cmd = AxiReadRespCmd(cmd.address, cmd.length, cmd.size, cycles, cmd.prot, burst_list, cmd.event)
             self.tag_context_manager.start_cmd(arid, resp_cmd)
@@ -925,27 +945,14 @@ class AxiMasterRead(Region, Reset):
             self.current_read_command = None
 
     async def _process_read_resp(self):
-        burst = []
-        cur_rid = None
-
         while True:
             r = await self.r_channel.recv()
 
             rid = int(getattr(r, 'rid', 0))
 
-            if cur_rid is not None and cur_rid != rid:
-                raise Exception(f"ID not constant within burst (expected {cur_rid}, got {rid})")
+            assert self.active_id[rid] > 0, "unexpected burst ID"
 
-            if self.active_id[rid] <= 0:
-                raise Exception(f"Unexpected burst ID {rid}")
-
-            burst.append(r)
-            cur_rid = rid
-
-            if int(r.rlast):
-                self.tag_context_manager.put_resp(rid, burst)
-                burst = []
-                cur_rid = None
+            self.tag_context_manager.put_resp(rid, r)
 
     async def _process_read_resp_id(self, context, cmd):
         rid = context.current_tag
@@ -966,14 +973,18 @@ class AxiMasterRead(Region, Reset):
         first = True
 
         for burst_length in cmd.burst_list:
-            burst = await context.get_resp()
+            for k in range(burst_length):
+                r = await context.get_resp()
 
-            if len(burst) != burst_length:
-                raise Exception(f"Burst length incorrect (ID {rid}, expected {burst_length}, got {len(burst)}")
+                assert self.active_id[rid] > 0, "unexpected burst ID"
+
+                if k == burst_length-1:
+                    assert int(r.rlast), "missing rlast at end of burst"
+                else:
+                    assert not int(r.rlast), "unexpected rlast within burst"
 
-            for r in burst:
                 cycle_data = int(r.rdata)
-                cycle_resp = AxiResp(getattr(r, "rresp", AxiResp.OKAY))
+                cycle_resp = AxiResp(int(getattr(r, "rresp", AxiResp.OKAY)))
                 cycle_user = int(getattr(r, "ruser", 0))
 
                 if cycle_resp != AxiResp.OKAY:
